/* BalooBluetooth
 * Author: Johan Ho
 */
using System;
using System.Net;
using System.IO;
using System.Threading;
using System.Windows.Forms;
using System.Text;
using System.Net.Sockets;
using System.Security.Cryptography;
using InTheHand.Net.Bluetooth;
using InTheHand.Net.Sockets;
using InTheHand.Net;

namespace Baloo
{
    /// <summary>
    /// This class is used to start a thread which listens to incoming Bluetooth
    /// requests and receives them. The requests are then forwarded to BalooBroadcast
    /// for processing.
    /// 
    /// There are actually two listeners, because apparently Windows Mobile can't send
    /// txt files to Bluetooth devices unless they are are listening with the File Transfer
    /// Profile, while other devices use the OBEX (OBjext EXchange) Push Profile.
    /// 
    /// So there's one listener using OBEX Push, and another using FTP, each in their
    /// own thread, and the data received is also parsed slightly differently, because
    /// the 32feet API includes classes for simplifying OBEX Push connections, but not
    /// for FTP connections.
    /// 
    /// The replying functions are the same, however.
    /// </summary>
    class BalooBluetooth
    {
        private BalooBroadcast broadcast;
        private BluetoothRadio radio;
        private ObexListener listener;
        private Thread listenthread;
        private Thread listenthreadftp;
        private ListView threadviewer;
        private int requestcounter = 1;
        private BluetoothListener blistener;
        internal static string salt;

        public BalooBluetooth(BalooBroadcast broadcast, ListView threadviewer)
        {
            this.broadcast = broadcast;
            this.threadviewer = threadviewer;
            this.radio = BluetoothRadio.PrimaryRadio;
            RNGCryptoServiceProvider rng = new RNGCryptoServiceProvider();
            byte[] saltbytes = new byte[4];
            rng.GetBytes(saltbytes);
            salt = Encoding.UTF8.GetString(saltbytes);
        }

        /// <summary>
        /// Start listening to incoming Bluetooth connections. This also sets the
        /// underlying BalooBroadcast object to the started state.
        /// </summary>
        /// <returns>True if broadcasting is started successfully, false if no Bluetooth radios are found.</returns>
        public bool StartBroadcast()
        {
            if (BluetoothRadio.AllRadios.Length == 0)
            {
                broadcast.LogEvent("Error starting broadcast: No Bluetooth radios found!");
                return false;
            }
            BluetoothRadio.PrimaryRadio.Mode = RadioMode.Discoverable;
            broadcast.StartBroadcast();
            
            listener = new ObexListener(ObexTransport.Bluetooth);
            listener.Start();
            listenthread = new Thread(new ThreadStart(ListenIncomingRequests));
            listenthread.Start();

            blistener = new BluetoothListener(BluetoothService.ObexFileTransfer);
            blistener.Start();
            listenthreadftp = new Thread(new ThreadStart(ListenIncomingRequestsFTP));
            listenthreadftp.Start();
            return true;
        }

        /// <summary>
        /// Stops listening. Also resets the BalooBroadcast object.
        /// </summary>
        public void StopBroadcast()
        {
            try
            {
                blistener.Stop();
                listener.Stop();
                listenthread.Abort();
                listenthreadftp.Abort();
            }
            catch
            {
            }
            listener = null;
            blistener = null;
            broadcast.StopBroadcast();
        }

        /// <summary>
        /// This function is started in it's own thread, to listen to incoming
        /// requests without "hanging" the GUI.
        /// </summary>
        private void ListenIncomingRequests()
        {
            while (listener != null && listener.IsListening)
            {
                try
                {
                    ObexListenerContext context = listener.GetContext();
                    if (!listener.IsListening)
                        broadcast.LogEvent("Listener stopped listening (probably because broadcast was stopped)");
                    else if (context == null)
                        broadcast.LogEvent("Can't get context (API error?)");
                    else
                    {
                        broadcast.LogEvent("Incoming connection");
                        // Start a new thread for handling the reply, before going back to listening.
                        ObexListenerRequest request = context.Request;
                        BalooRequestHandler handler = new BalooRequestHandler(broadcast,threadviewer,requestcounter++,request);
                        Thread handlerthread = new Thread(new ThreadStart(handler.HandleIncomingRequest));
                        handlerthread.Start();
                    }
                }
                catch //(Exception e)
                {
                    //Enable this when debugging.
                    //MessageBox.Show(e.ToString());
                }
            }
        }

        /// <summary>
        /// Listens to incoming FTP requests without "hanging" the GUI.
        /// </summary>
        private void ListenIncomingRequestsFTP()
        {
            while (blistener != null)
            {
                try
                {
                    BluetoothClient client = blistener.AcceptBluetoothClient();
                    broadcast.LogEvent("Incoming FTP connection");
                    BalooRequestHandler handler = new BalooRequestHandler(broadcast, threadviewer, requestcounter++, client);
                    Thread handlerthread = new Thread(new ThreadStart(handler.HandleIncomingClient));
                    handlerthread.Start();
                }
                catch //(Exception e)
                {
                    //Enable this when debugging.
                    //MessageBox.Show(e.ToString());
                }
            }
        }


        /// <summary>
        /// This class is used for starting a new thread for handling requests/replies.
        /// (Since threads don't take arguments, a class is needed to store the arguments
        /// for the new threads.)
        /// </summary>
        private class BalooRequestHandler
        {
            private ObexListenerRequest request;
            private BalooBroadcast broadcast;
            private ListView threadviewer;
            private ListViewItem listitem;
            private int requestid;
            private UpdateThreadStatusDelegate updatethreadstatus;
            private BluetoothClient client;

            private static int DEFAULTPACKETSIZE = 4096;
            // We insert a delay before sending a reply to the user, because their Bluetooth
            // connection may not have reset yet after sending the command/file to Baloo.
            private static int DELAYBEFOREREPLY = 5000;
            /* It seems that if the code doesn't delay between writing and reading from the
             * network stream, the connection will fail. Therefore a delay is put after
             * each write action. */
            private static int DELAYAFTERWRITE = 1000;
            private static int NUMRETRIES = 2;

            #region OBEX constants

            private static byte OBEXVERSION = 0x10;
            private static byte OBEXCONNECTFLAGS = 0x00;

            // OBEX packet type codes
            private static byte CONNECT = 0x80;
            private static byte DISCONNECT = 0x81;
            private static byte PUT = 0x02;
            private static byte PUTFINAL = 0x82;

            // OBEX header codes
            private static byte OBEXCONNECTIONID = 0xCB;
            private static byte OBEXNAME= 0x01;
            private static byte OBEXBODY = 0x48;
            private static byte OBEXBODYEND = 0x49;

            // OBEX response codes
            private static byte OBEXCONTINUE = 0x90;
            private static byte OBEXSUCCESS = 0xA0;

            // OBEX responses
            // 7 is the length of the byte array, 15,215 is the max packet size's hi and low byte (4040).
            private static byte[] ACCEPTCONNECTION = new byte[] { OBEXSUCCESS, 0, 7, OBEXVERSION, OBEXCONNECTFLAGS, 15, 215 };
            private static byte[] CONTINUESTRING = new byte[] { OBEXCONTINUE, 0, 3 };
            private static byte[] SUCCESSSTRING = new byte[] { OBEXSUCCESS, 0, 3 };
            #endregion

            /// <summary>
            /// Initializes the request handler with the request that is to be handled.
            /// </summary>
            /// <param name="broadcast">The BalooBroadcast instance, this is needed for
            /// running commands that are received.</param>
            /// <param name="threadviewer">The ListView component which shows info about the reply
            /// threads. (The threads that this class is used for.)</param>
            /// <param name="requestid">An id gotten from an internal counter in BalooBluetooth.
            /// Used for differentiating between requests in the debug log.</param>
            /// <param name="request">The Bluetooth request that this handler should take care of.</param>
            public BalooRequestHandler(BalooBroadcast broadcast, ListView threadviewer, int requestid, ObexListenerRequest request)
            {
                this.broadcast = broadcast;
                this.threadviewer = threadviewer;
                this.requestid = requestid;
                this.listitem = new ListViewItem(new string[] { requestid.ToString(), "Received" });
                threadviewer.Invoke(new ListViewAdd(threadviewer.Items.Add), new object[] { this.listitem });
                updatethreadstatus = new UpdateThreadStatusDelegate(UpdateThreadStatus);
                this.request = request;
            }

            /// <summary>
            /// A request handler which handles requests from the FTP profile instead of OBEX
            /// profile.
            /// </summary>
            /// <param name="broadcast">The BalooBroadcast instance, this is needed for
            /// running commands that are received.</param>
            /// <param name="threadviewer">The ListView component which shows info about the reply
            /// threads. (The threads that this class is used for.)</param>
            /// <param name="requestid">An id gotten from an internal counter in BalooBluetooth.
            /// Used for differentiating between requests in the debug log.</param>
            /// <param name="client">The client gotten from the FTP listener.</param>
            public BalooRequestHandler(BalooBroadcast broadcast, ListView threadviewer, int requestid, BluetoothClient client)
            {
                this.broadcast = broadcast;
                this.threadviewer = threadviewer;
                this.requestid = requestid;
                this.listitem = new ListViewItem(new string[] { requestid.ToString(), "Received" });
                threadviewer.Invoke(new ListViewAdd(threadviewer.Items.Add), new object[] { this.listitem });
                updatethreadstatus = new UpdateThreadStatusDelegate(UpdateThreadStatus);
                this.client = client;
            }

            #region Functions for updating the listview
            /// <summary>Delegate for adding an item to the ListView Items collection.</summary>
            public delegate ListViewItem ListViewAdd(ListViewItem item);
            /// <summary>Delegate for removing an item from the ListView Items collection.</summary>
            public delegate void ListViewRemove(ListViewItem item);
            /// <summary>Delegate for updating the status text of the ListViewItem.</summary>
            public delegate void UpdateThreadStatusDelegate(string message);
            /// <summary>
            /// Updates the ListViewItem's status text. (Must be used with the delegate, because
            /// this is a different thread!)
            /// </summary>
            /// <param name="text">The new status text</param>
            public void UpdateThreadStatus(string text)
            {
                listitem.SubItems[1].Text = text;
            }
            #endregion

            #region Request handling
            /// <summary>
            /// This is a helper function for handling a request that has been received
            /// by the ListenIncomingRequests() function/thread.
            /// </summary>
            public void HandleIncomingRequest()
            {
                try
                {
                    WebHeaderCollection headers = request.Headers;
                    BluetoothEndPoint clientep = (BluetoothEndPoint)request.RemoteEndPoint;

                    /* Log data about the incoming request. This data can be useful
                     * in case of errors, to see what kind of request caused the errors. */
                    //broadcast.LogEvent("Received from: " + clientep.Address.ToString()); // Don't log address for user privacy
                    LogWithRequestID("Protocol version: " + request.ProtocolVersion);
                    LogWithRequestID("Obex method: " + request.ObexMethod);
                    LogWithRequestID("Raw URL: " + request.RawUrl);
                    LogWithRequestID("URL: " + request.Url);
                    foreach (string key in headers.AllKeys)
                        LogWithRequestID("Header: " + key + " => " + headers.Get(key));

                    // If the received request is a text file, interpret it as a command.
                    if ((headers.Get("NAME")).EndsWith(".txt"))
                    {
                        threadviewer.Invoke(updatethreadstatus,new object[] { "Running command..." });
                        // Read the command from the request stream
                        StreamReader reader = new StreamReader(request.InputStream);
                        string commandstring = reader.ReadToEnd();
                        reader.Close();
                        // Remove null characters from end of string
                        int nullpos = commandstring.IndexOf('\0');
                        if (nullpos > 0)
                            commandstring = commandstring.Substring(0, commandstring.IndexOf('\0'));
                        // If position is 0, it means the string consists of only null characters
                        else if (nullpos == 0)
                            commandstring = "";
                        // Change lineshift characters to space, and trim space from beginning and end
                        commandstring = commandstring.Replace('\n', ' ').Replace('\r', ' ').Trim();
                        commandstring = commandstring.Trim();
                        if (commandstring.Equals(""))
                            LogWithRequestID("No commands in request!");
                        else
                        {
                            BalooReply reply = broadcast.RunCommand(commandstring, GetHash(clientep.Address.ToString()) );
                            HandleReply(reply, clientep);
                        }
                    }
                    // Otherwise, handle it as a non-text file.
                    else
                    {
                        threadviewer.Invoke(updatethreadstatus,new object[] { "Handling incoming file..." });
                        HandleFile(request);
                    }
                    LogWithRequestID("Finished!" + Environment.NewLine);
                    threadviewer.Invoke(
                        new ListViewRemove(threadviewer.Items.Remove),
                        new object[] { this.listitem }
                    );
                }
                // If an exception occurs, make sure to remove the request/thread from the list
                catch (Exception e)
                {
                    try
                    {
                        threadviewer.Invoke(
                            new ListViewRemove(threadviewer.Items.Remove),
                            new object[] { this.listitem }
                        );
                        LogWithRequestID(e.ToString());
                        MessageBox.Show(e.ToString());
                    }
                    // If an exception occurs here, it's probably because Baloo is closed before
                    // the broadcast is completely stopped.
                    catch { }
                }
            }

            /// <summary>
            /// Helper function for creating a hash code using MD5 and the randomly selected salt.
            /// </summary>
            /// <param name="text"></param>
            /// <returns></returns>
            private int GetHash(string text)
            {
                byte[] textbytes = Encoding.ASCII.GetBytes(salt+text);
                HashAlgorithm hasher = new MD5CryptoServiceProvider();
                byte[] hashresult = hasher.ComputeHash(textbytes);
                return Encoding.UTF8.GetString(hashresult).GetHashCode();
            }

            /// <summary>
            /// This function handles incoming files (that are not text files).
            /// </summary>
            /// <param name="request">The incoming request.</param>
            private void HandleFile(ObexListenerRequest request)
            {
                BluetoothEndPoint ep = (BluetoothEndPoint)request.RemoteEndPoint;
                BalooReply reply = broadcast.HandleIncomingFile(request.Headers.Get("NAME"), GetHash(ep.Address.ToString()) );

                if (reply.Filename != null)
                {
                    threadviewer.Invoke(updatethreadstatus, new object[] { "Saving file..." });
                    request.WriteFile(reply.Filename);
                    LogWithRequestID("File saved to " + reply.Filename);
                } else {
                    LogWithRequestID("File not accepted");
                }
                HandleReply(reply, ep);
            }
            #endregion

            #region File Transfer Profile request handling
            /// <summary>
            /// Helper function for logging bytes shown as bytes.
            /// This is used mainly for debugging, and is partly disabled because it takes a lot
            /// of resources if it tries to show all the bytes from a large received file.
            /// I've left logging of number of bytes.
            /// </summary>
            /// <param name="buffer">The buffer of bytes to show.</param>
            /// <param name="length">The number of bytes to show from the buffer.</param>
            /// <param name="description">A description of the bytes that are shown.</param>
            private void LogBytes(byte[] buffer, int length, string description)
            {
                string bytes = "(disabled)";
                /*for (int index = 0; index < length; index++)
                    bytes += String.Format("0x{0:x2},", buffer[index]);*/
                LogWithRequestID(description + " ("+length+"b): " + bytes);
            }

            /// <summary>
            /// Helper function for adding request id in front of the log text.
            /// </summary>
            /// <param name="logtext">The logtext to be logged.</param>
            private void LogWithRequestID(string logtext)
            {
                broadcast.LogEvent("Request ID " + requestid + " - " + logtext);
            }

            /// <summary>
            /// Helper function for writing to a stream, logging this and delaying after the write.
            /// </summary>
            /// <param name="stream">The stream to be written to.</param>
            /// <param name="buffer">The buffer to be written.</param>
            /// <param name="logstring">The text to be logged.</param>
            private void StreamWrite(NetworkStream stream, byte[] buffer, string logstring)
            {
                LogWithRequestID(logstring);
                stream.Write(buffer, 0, buffer.Length);
                Thread.Sleep(DELAYAFTERWRITE);
            }

            /// <summary>
            /// Handles a BluetoothClient object, reading the data stream from it and fetches
            /// parameters from the data read. Then uses it to handle the request.
            /// 
            /// This code was mostly made from trial and failure, and may still have bugs!
            /// </summary>
            public void HandleIncomingClient()
            {
                BluetoothEndPoint clientep = (BluetoothEndPoint)client.Client.RemoteEndPoint;
                string filename = null;
                byte[] body = null;

                try
                {
                    NetworkStream stream = client.GetStream();
                    byte[] buffer = new byte[DEFAULTPACKETSIZE];

                    // Read connect string, and see if it is valid (begins with connect command).
                    int bytesread = stream.Read(buffer, 0, buffer.Length);
                    LogBytes(buffer, bytesread, "Connection string");
                    if (buffer[0] == CONNECT)
                    {
                        // Return connect acception code
                        StreamWrite(stream, ACCEPTCONNECTION, "Send accept connection code");
                        /* The max packet size should be defined in byte 5 and 6, but we make sure
                         * that it's larger than the default packet size before setting it, in case
                         * it for some reason is not set in those bytes. */
                        int maxpacketsize = buffer[5] * 0x0100 + buffer[6];
                        if (maxpacketsize > DEFAULTPACKETSIZE)
                            buffer = new byte[maxpacketsize];
                    }
                    else return; // Not a connection?

                    bytesread = stream.Read(buffer, 0, buffer.Length); // Read packet
                    // Loop for reading and writing to the network stream untill no more bytes are
                    // read from the stream.
                    while (bytesread > 0)
                    {
                        LogWithRequestID("OBEX packet type: " + buffer[0]);
                        LogBytes(buffer, bytesread, "Received data");

                        if (buffer[0] == PUT || buffer[0] == PUTFINAL)
                        {
                            int index = 3;
                            /* If the first header is connection id, skip past it.
                             * It can't be treated in the same way as other headers because
                             * it doesn't specify length, and instead of adding a check for
                             * it, we just skip past it since it isn't used. */
                            if (buffer[index] == OBEXCONNECTIONID)
                                index += 5;

                            // Loop through the bytes to fetch headers
                            // (though most headers are ignored)
                            while (index < bytesread)
                            {
                                int header = buffer[index];
                                LogWithRequestID(String.Format("Index: {0}", index) + String.Format(", Header type: 0x{0:x2}", header));
                                int len = buffer[index + 1] * 0x0100 + buffer[index + 2] - 3;
                                // It's -3 because the first three bytes are counted in the length
                                int toread = len;
                                LogWithRequestID("Header length: " + len);
                                byte[] headerbuffer = new byte[len];
                                index += 3;

                                // If the length of the header goes past the size of the buffer, it means
                                // we have to fetch more packets to get the whole header.
                                while (toread > bytesread - index)
                                {
                                    LogWithRequestID("Must fetch more packets");
                                    Array.Copy(buffer, index, headerbuffer, len - toread, bytesread - index);
                                    toread -= bytesread - index;
                                    StreamWrite(stream, CONTINUESTRING, "Send continue code");
                                    bytesread = stream.Read(buffer, 0, buffer.Length);
                                    LogBytes(buffer, bytesread, "Received data");
                                    index = 0;
                                }
                                // Read remaining bytes (if the previous loop wasn't used, this would
                                // just read normally from the current packet).
                                if (toread > 0)
                                {
                                    Array.Copy(buffer, index, headerbuffer, len - toread, toread);
                                    // Skip past the bytes that were read.
                                    index += toread;
                                }
                                LogBytes(headerbuffer, len, "Header contents");

                                // If this header is the obexname header, it should be the filename of the
                                // sent file.
                                if (header == OBEXNAME)
                                {
                                    // The filename is closed with two null characters, so ignore these
                                    // when making the string.
                                    filename = Encoding.BigEndianUnicode.GetString(headerbuffer, 0, len - 2);
                                    LogWithRequestID("Filename: " + filename);
                                    // Skip past the closing bytes for the filename
                                    index += 5;
                                }

                                // If header is body, it means it's the data body of the sent file, and it
                                // should therefore stored in the body buffer.
                                else if ((header == OBEXBODY || header == OBEXBODYEND) && len > 0)
                                {
                                    if (body == null)
                                        body = headerbuffer;
                                    // If part of body is already read, append the new bodypart at the end.
                                    else
                                    {
                                        byte[] newbody = new byte[body.Length + headerbuffer.Length];
                                        Array.Copy(body, newbody, body.Length);
                                        Array.Copy(headerbuffer, 0, newbody, body.Length, headerbuffer.Length);
                                        body = newbody;
                                    }
                                }
                            }
                            // If putfinal, send success code
                            if (buffer[0] == PUTFINAL)
                                StreamWrite(stream, SUCCESSSTRING, "Send success code");
                            // If not, assume there's more to fetch
                            else StreamWrite(stream, CONTINUESTRING, "Send continue code");
                        }
                        // For all other OBEX packets, send success code untill it disconnects
                        else StreamWrite(stream, SUCCESSSTRING, "Send success code");
                        bytesread = stream.Read(buffer, 0, buffer.Length);
                    }
                }
                catch (Exception e)
                {
                    LogWithRequestID("Error receiving data: " + e.ToString());
                }
                // If data has been received properly, handle it
                if (filename != null && body != null)
                {
                    // Parse data as command if it's a txt file.
                    if (filename.EndsWith(".txt"))
                    {
                        threadviewer.Invoke(updatethreadstatus, new object[] { "Running command..." });
                        string commandstring = Encoding.GetEncoding("ISO-8859-1").GetString(body).Trim();
                        LogWithRequestID("Commandstring: " + commandstring);
                        BalooReply reply = broadcast.RunCommand(commandstring, GetHash(clientep.Address.ToString()) );
                        if (reply != null)
                            HandleReply(reply, clientep);

                        // Otherwise, check if the file should be saved.
                    }
                    else HandleFileFTP(clientep, filename, body);
                }
                // If data not received correctly, log proper error messages.
                else
                {
                    if (filename == null)
                        LogWithRequestID("Filename not gotten!");
                    else LogWithRequestID("Filename gotten, but no data received!");
                    HandleReply(new BalooReply( BalooReply.Types.Text,
                        broadcast.settings.IncomingFileNotAccepted),clientep);
                }

                threadviewer.Invoke(updatethreadstatus, new object[] { "Finished handling!" });
                LogWithRequestID("Finished!" + Environment.NewLine);
                threadviewer.Invoke(
                    new ListViewRemove(threadviewer.Items.Remove),
                    new object[] { this.listitem }
                );
            }

            /// <summary>
            /// Helper file handling function for FTP listener.
            /// </summary>
            /// <param name="clientep">The client which should receive the response</param>
            /// <param name="filename">The filename of the received file.</param>
            /// <param name="body">The file data.</param>
            private void HandleFileFTP(BluetoothEndPoint clientep, string filename, byte[] body)
            {
                BalooReply reply = broadcast.HandleIncomingFile(filename, GetHash(clientep.Address.ToString()) );
                if (reply.Filename != null)
                {
                    threadviewer.Invoke(updatethreadstatus, new object[] { "Saving file..." });
                    try
                    {
                        FileStream stream = File.OpenWrite(reply.Filename);
                        stream.Write(body, 0, body.Length);
                        stream.Close();
                        LogWithRequestID("File saved to " + reply.Filename);
                    }
                    catch (Exception e)
                    {
                        reply.Text = broadcast.settings.IncomingFileNotAccepted;
                        LogWithRequestID("Error saving file: " + e.Message);
                    }
                }
                else
                {
                    LogWithRequestID("File not accepted");
                }
                HandleReply(reply, clientep);
            }
            #endregion

            #region Replying functions (outgoing connections)
            /// <summary>
            /// Handles replying to the user, whether the reply is a string or a file.
            /// </summary>
            /// <param name="reply">The reply data.</param>
            /// <param name="recipientEndPoint">The recipient of the reply.</param>
            private void HandleReply(BalooReply reply, BluetoothEndPoint recipientEndPoint)
            {
                int tryagain = NUMRETRIES;
                while (tryagain > 0)
                {
                    tryagain--;
                    threadviewer.Invoke(updatethreadstatus, new object[] { "Delaying reply..." });
                    ObexWebResponse response = null;
                    // Sleeps a bit to make sure Bluetooth is usable on the client device after it
                    // sent the command.
                    Thread.Sleep(DELAYBEFOREREPLY);
                    threadviewer.Invoke(updatethreadstatus, new object[] { "Sending reply..." });
                    if (reply.Type == BalooReply.Types.Text)
                        response = SendText(recipientEndPoint.Address, reply.Text, reply.Filename);
                    else if (reply.Type == BalooReply.Types.Filename)
                        response = SendFile(recipientEndPoint.Address, reply.Filename);
                    else if (reply.Type == BalooReply.Types.Filebuffer)
                        response = SendBuffer(recipientEndPoint.Address, reply.Filename, reply.Databuffer);
                    if (response != null)
                    {
                        LogWithRequestID("Response code: " + response.StatusCode.ToString() );
                        if ((response.StatusCode & ObexStatusCode.Success) == ObexStatusCode.Success)
                        {
                            LogWithRequestID("Reply sent!");
                            threadviewer.Invoke(updatethreadstatus, new object[] { "Reply sent!" });
                            tryagain = 0;
                        }
                        else if ((response.StatusCode & ObexStatusCode.BadRequest) == ObexStatusCode.BadRequest)
                        {
                            string retrystring = "";
                            if (tryagain > 0)
                                retrystring = " Retrying...";
                            LogWithRequestID("Error sending." + retrystring);
                            threadviewer.Invoke(updatethreadstatus, new object[] { "Error - bad request." + retrystring });
                        }
                        else
                        {
                            LogWithRequestID("Unexpected response, cancelling.");
                            threadviewer.Invoke(updatethreadstatus, new object[] { "Unexpected response, cancelling." });
                            tryagain = 0;
                        }
                        response.Close();
                    }
                    else
                    {
                        LogWithRequestID("Unrecognized error sending reply.");
                        threadviewer.Invoke(updatethreadstatus, new object[] { "Error sending reply." });
                        tryagain = 0;
                    }
                }
            }

            /// <summary>
            /// Sends a text message to the user.
            /// </summary>
            /// <param name="recipient">The recipient of the message.</param>
            /// <param name="text">The message to be sent.</param>
            /// <param name="filename">The filename to send the message with. Can be null,
            /// in which case Balooreply.txt is used by default.</param>
            /// <returns>The response object of the sending action.</returns>
            private ObexWebResponse SendText(BluetoothAddress recipient, string text, string filename)
            {
                LogWithRequestID("Sending reply (filename: " + filename + "): " + text);
                if (String.IsNullOrEmpty(filename))
                    filename = "Balooreply.txt";
                Uri uri = new Uri("obex://" + recipient.ToString() + "/" + filename);
                //broadcast.LogEvent(uri.ToString());
                ObexWebRequest request = new ObexWebRequest(uri);
                // Convert to ISO-8859-1 for  to work properly on non-unicode devices.
                StreamWriter writer = new StreamWriter(request.GetRequestStream(), Encoding.GetEncoding("ISO-8859-1"));
                writer.Write(text);
                writer.Close();
                return (ObexWebResponse)request.GetResponse();
            }

            /// <summary>
            /// Sends a file to the user.
            /// </summary>
            /// <param name="recipient">The recipient of the file.</param>
            /// <param name="filepath">The full path to the file to be sent.</param>
            /// <returns>The response object of the sending action.</returns>
            private ObexWebResponse SendFile(BluetoothAddress recipient, string filepath)
            {
                LogWithRequestID("Sending file: " + filepath);
                System.Uri uri = new Uri("obex://" + recipient.ToString() + "/" + System.IO.Path.GetFileName(filepath));

                ObexWebRequest request = new ObexWebRequest(uri);
                request.ReadFile(filepath);
                return (ObexWebResponse)request.GetResponse();
            }

            /// <summary>
            /// Sends a binary file buffer to the user.
            /// </summary>
            /// <param name="recipient">The recipient of the message.</param>
            /// <param name="filename">The filename to send the data as.</param>
            /// <returns>The buffer with the data to send.</returns>
            private ObexWebResponse SendBuffer(BluetoothAddress recipient, string filename, byte[] buffer)
            {
                LogWithRequestID("Sending file: " + filename);
                if (filename == null)
                    filename = "Balooreply.txt";
                Uri uri = new Uri("obex://" + recipient.ToString() + "/" + filename);
                ObexWebRequest request = new ObexWebRequest(uri);
                Stream stream = request.GetRequestStream();
                stream.Write(buffer, 0, buffer.Length);
                stream.Close();
                return (ObexWebResponse)request.GetResponse();
            }
            #endregion
        }
    }
}
