/* BalooBroadcast
 * Author: Johan Ho
 */
using System;
using System.IO;
using System.Collections;
using System.Text;
using System.Windows.Forms;

namespace Baloo
{
    /// <summary>
    /// This class manages requests and replies to them via BalooBluetooth.
    /// </summary>
    class BalooBroadcast
    {
        internal BalooSettings settings;
        private TextBox logbox;
        private BalooHTTP http;
        private BalooLogging log;
        private BalooStatistics stats;
        private ArrayList incomingfilter;
        // The following string is used for storing the command list, so it doesn't
        // need to be regenerated every time it's requested.
        private string commandlist;
        private bool broadcasting = false;

        /// <summary>
        /// Initializes this instance with a new BalooHTTP object. Since the BalooHTTP
        /// class doesn't need any parameters, it doesn't need to be recreated even if
        /// the configuration changes.
        /// </summary>
        /// <param name="logbox">The TextBox which should be used for logging.</param>
        public BalooBroadcast( TextBox logbox )
        {
            this.logbox = logbox;
            http = new BalooHTTP();
        }

        /// <summary>
        /// Starts the broadcast:
        /// Initializes new instances of BalooLogging and BalooStatistics with the
        /// current settings for logging and statistics.
        /// Also generates the command list from the current collection of resources.
        /// </summary>
        public void StartBroadcast()
        {
            log = new BalooLogging(settings.LogFile);
            stats = new BalooStatistics(settings.MySQLServer,settings.MySQLUsername,settings.MySQLPassword,settings.MySQLDatabase);
            if (!stats.ConnectionOK())
                LogEvent("Warning: Connection to MySQL server was not OK, statistics may not be saved!");
            // Generate the commandlist.
            commandlist = "";
            foreach (BalooResource resource in settings.Resources)
                if (!resource.Hidden && !resource.Disabled && resource.ResourceType != BalooResource.ResourceTypes.CommandList)
                    commandlist += resource.Command + Environment.NewLine + resource.Description + Environment.NewLine + Environment.NewLine;
            incomingfilter = new ArrayList(settings.IncomingFilter.Split(new char[] { ',' }));
            broadcasting = true;
        }

        /// <summary>
        /// Stops the broadcast:
        /// Deletes the instances of BalooLogging and BalooStatistics.
        /// </summary>
        public void StopBroadcast()
        {
            log = null;
            stats = null;
            broadcasting = false;
        }

        #region Logging and statistics
        /// <summary>
        /// Logs debug text to the logfile and in the logbox in the GUI.
        /// A datestamp is added in front of the debug text.
        /// </summary>
        /// <param name="text">The text to be logged.</param>
        public void LogEvent(string text)
        {
            DateTime dt = DateTime.Now;
            text = dt.ToShortDateString() + " " + dt.ToLongTimeString() + ": " + text;
            if (logbox != null)
                logbox.Invoke(new LogBoxDelegate(LogBoxAddText), new object[] { text });
            if (broadcasting)
                log.LogEvent(text);
        }

        /// <summary>
        /// This delegate is used for calling the LogBoxAddText function from another
        /// thread. (The Bluetooth listening thread.)
        /// </summary>
        /// <param name="text">The text to be added.</param>
        public delegate void LogBoxDelegate(string text);

        /// <summary>
        /// Adds text to the logbox and scrolls it down to show the newest text.
        /// </summary>
        /// <param name="text">The text to be added.</param>
        private void LogBoxAddText(string text)
        {
            logbox.Text += text + Environment.NewLine;
            logbox.Select(logbox.Text.Length-1, 1);
            logbox.ScrollToCaret();
        }

        /// <summary>
        /// This function creates a temporary instance of BalooStatistics, then calls
        /// the GenerateStatistics function from it, before deleting the instance again.
        /// </summary>
        /// <param name="dataGridStatistics">The datagrid component to put the statistics in.</param>
        /// <param name="StatisticsTypes">The type of statistics requested.</param>
        public string GenerateStatistics(DataGridView dataGridStatistics, BalooStatistics.StatisticsTypes StatisticsTypes)
        {
            string result;
            if (stats == null)
            {
                stats = new BalooStatistics(settings.MySQLServer, settings.MySQLUsername, settings.MySQLPassword, settings.MySQLDatabase);
                result = stats.GenerateStatistics(dataGridStatistics, StatisticsTypes);
                stats = null;
            }
            else result = stats.GenerateStatistics(dataGridStatistics, StatisticsTypes);
            return result;
        }

        public int DeleteEventsBefore(DateTime date)
        {
            int result;
            if (stats == null)
            {
                stats = new BalooStatistics(settings.MySQLServer, settings.MySQLUsername, settings.MySQLPassword, settings.MySQLDatabase);
                result = stats.DeleteEventsBefore(date);
                stats = null;
            }
            else result = stats.DeleteEventsBefore(date);
            return result;
        }
        #endregion

        #region Handling requests
        /// <summary>
        /// Tries to "run" the supplied commandstring.
        /// </summary>
        /// <param name="commandstring">The commandstring (command plus eventual parameters.</param>
        /// <param name="userhash">A hash of the user's Bluetooth ID, used for calculating unique hits.</param>
        /// <returns>The reply data.</returns>
        public BalooReply RunCommand(string commandstring,int userhash)
        {
            if (commandstring == null || commandstring.Length == 0)
                return null;
            // Split the commandstring into command and parameters, if there are any parameters.
            LogEvent("Commandstring: \"" + commandstring + "\"" );
            int spacepos = commandstring.IndexOf(' ');
            string command, parameters=null;
            if (spacepos > 0)
            {
                command = commandstring.Substring(0, spacepos);
                parameters = commandstring.Substring(spacepos + 1);
            }
            else command = commandstring;
            LogEvent("Command: \"" + command + "\"");
            if (parameters != null)
                LogEvent("Parameters: \"" + parameters + "\"");

            // Go through the resources to find the resource that is called by the supplied command.
            BalooReply reply = null;
            foreach (BalooResource resource in settings.Resources)
                if (!resource.Disabled && resource.Command != null && String.Compare(resource.Command, command,true)==0)
                {
                    switch (resource.ResourceType)
                    {
                        case BalooResource.ResourceTypes.CommandList:
                            reply = new BalooReply( BalooReply.Types.Text,
                                resource.Description + Environment.NewLine +
                                Environment.NewLine + commandlist);
                            break;
                        case BalooResource.ResourceTypes.File:
                            if (File.Exists(resource.Location))
                                reply = new BalooReply(BalooReply.Types.Filename,resource.Location);
                            else reply = new BalooReply(BalooReply.Types.Text,settings.FileNotFound);
                            break;
                        case BalooResource.ResourceTypes.Folder:
                            reply = RunFolderCommand(parameters, resource);
                            break;
                        case BalooResource.ResourceTypes.URL:
                            reply = http.FetchURL(resource.Location, parameters);
                            break;
                    }
                    if (reply != null)
                    {
                        if (reply.Type == BalooReply.Types.Text)
                            stats.AddEvent(command, parameters, reply.Text, userhash);
                        else stats.AddEvent(command, parameters, "(File: " + reply.Filename + ")", userhash);
                        return reply;
                    }
                }
            // If the code arrives here, it means no resource matching the given command was found.
            reply = new BalooReply( BalooReply.Types.Text, settings.ResourceUnavailable +
                Environment.NewLine + Environment.NewLine + commandlist );
            if (reply.Type == BalooReply.Types.Text)
                stats.AddEvent(command, parameters, reply.Text, userhash);
            else stats.AddEvent(command, parameters, "(File: " + reply.Filename + ")", userhash);
            return reply;
        }

        /// <summary>
        /// This is a helper function for RunCommand, and decides what to do with a
        /// command calling a folder resource, depending on the given parameters.
        /// </summary>
        /// <param name="parameters">The parameters that were given with the command</param>
        /// <param name="resource">The resource that is requested.</param>
        /// <returns>The reply data.</returns>
        private BalooReply RunFolderCommand(string parameters, BalooResource resource)
        {
            // If the directory specified by the object doesn't exist, nothing can
            // be done.
            if (!Directory.Exists(resource.Location))
                return new BalooReply( BalooReply.Types.Text, settings.ResourceUnavailable);
            // There are several paths which require a file listing, so we use a bool
            // to specify which paths need it.
            bool filelisting = false;

            BalooReply reply = null;
            // If no parameters are given, just give the description of the folder and
            // a file listing of the folder.
            if (parameters == null)
            {
                reply = new BalooReply(BalooReply.Types.Text,
                    resource.Description + Environment.NewLine + Environment.NewLine);
                filelisting = true;
            }
            // If parameters are given, attempt to return the file that is requested.
            else
            {
                string getfile = resource.Location + "\\" + parameters;
                // If file is found, the response should be the file.
                if (File.Exists(getfile))
                    reply = new BalooReply( BalooReply.Types.Filename,getfile);
                // If not, the response is an error message and a file listing.
                else
                {
                    reply = new BalooReply( BalooReply.Types.Text,
                        settings.FileNotFound + Environment.NewLine + Environment.NewLine);
                    filelisting = true;
                }
            }
            // If file listing is needed, add it to the current response text.
            if (filelisting)
            {
                string[] filenames = Directory.GetFiles(resource.Location);
                foreach (string filename in filenames)
                    reply.Text += filename.Substring(filename.LastIndexOf('\\')+1) + Environment.NewLine;
            }
            return reply;
        }

        /// <summary>
        /// This function handles an incoming nontext file, and decides whether to
        /// save it or not. It returns the full path + file name if the file should
        /// be saved, or null if it should not be saved.
        /// </summary>
        /// <param name="filename">The file name of the incoming file.</param>
        /// <param name="hash">A hash of the user's Bluetooth ID.</param>
        /// <returns>The reply data, should be of type Text, but with Filename set as
        /// full path + file name if the file should be saved, or null if it should not.</returns>
        public BalooReply HandleIncomingFile(string filename, int hash)
        {
            BalooReply reply = new BalooReply(BalooReply.Types.Text, null);
            // Check if file type is filtered first.
            int dotpos = filename.IndexOf('.');
            if (dotpos >= 0)
            {
                string filetype = filename.Substring(dotpos);
                if (incomingfilter.Contains(filetype))
                {
                    reply.Text = settings.IncomingFileNotAccepted;
                    return null;
                }
            }
            // If a folder is set and exists, find a nonused file name for the incoming
            // file, and return the full path + file name.
            string folder = settings.SaveIncomingToFolder;
            if (folder != null && Directory.Exists(folder))
            {
                string fullpath = folder + "\\" + filename;
                int count = 1, dotindex;
                while (File.Exists(fullpath))
                {
                    dotindex = fullpath.LastIndexOf('.');
                    if (!File.Exists(fullpath.Substring(0, dotindex) + count + fullpath.Substring(dotindex)))
                        fullpath = fullpath.Substring(0, dotindex) + count + fullpath.Substring(dotindex);
                    count++;
                }
                reply.Text = settings.IncomingFileSaved;
                stats.AddEvent("(file received)", "Saved to: " + fullpath, settings.IncomingFileSaved, hash);
                reply.Filename = fullpath;

            // If not, return null and set the reply to an error message.
            } else {
                reply.Text = settings.IncomingFileNotAccepted;
                stats.AddEvent("(file received)", "Denied: " + filename, settings.IncomingFileNotAccepted, hash);
            }
            return reply;
        }
        #endregion
    }
}