/* BalooGUI.Resources
 * Author: Johan Ho
 * 
 * This partial class contains the code for handling user input on the GUI
 * components related to resources.
 */
using System;
using System.Windows.Forms;

namespace Baloo
{
    partial class BalooGUI
    {
        /// <summary>
        /// This helper function refreshes the resources listbox and is used when
        /// the resources are updated (name changed, new resource added or resource
        /// deleted).
        /// </summary>
        private void ResourcesUpdate()
        {
            Object selected = lstResources.SelectedItem;
            broadcast.settings.Resources.Sort();
            ((CurrencyManager)this.BindingContext[broadcast.settings.Resources]).Refresh();
            if (broadcast.settings.Resources.Count == 0)
                grpEditResource.Visible = false;
            
            if (selected != null && lstResources.Items.Contains(selected))
                lstResources.SelectedItem = selected;
            else lstResources.SelectedItem = null;
        }

        #region Adding, selecting, saving and deleting resources
        /// <summary>
        /// Adds a new BalooResource object to the resources, and updates
        /// the resources listbox.
        /// </summary>
        private void btnNewResource_Click(object sender, EventArgs e)
        {
            BalooResource res = new BalooResource();
            broadcast.settings.Resources.Add(res);
            ResourcesUpdate();
            lstResources.SelectedItem = null;
            lstResources.SelectedItem = res;
        }

        /// <summary>
        /// Shows the information for the selected resource.
        /// </summary>
        private void lstResources_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (lstResources.SelectedItem != null)
            {
                BalooResource resource = (BalooResource)lstResources.SelectedItem;
                SetTextBinding( txtResName, resource, "Name");
                SetTextBinding( txtResCommand, resource, "Command" );
                SetTextBinding( txtResDescription, resource, "Description" );
                if (resource.ResourceType != BalooResource.ResourceTypes.File)
                {
                    txtResFile.DataBindings.Clear();
                    txtResFile.Text = "";
                } else {
                    txtResFile.Text = resource.Location;
                    SetTextBinding(txtResFile, resource, "Location");
                    radioResFile.Checked = true;
                }
                if (resource.ResourceType != BalooResource.ResourceTypes.Folder)
                {
                    txtResFolder.DataBindings.Clear();
                    txtResFolder.Text = "";
                } else {
                    txtResFolder.Text = resource.Location;
                    SetTextBinding(txtResFolder, resource, "Location");
                    radioResFolder.Checked = true;
                }
                if (resource.ResourceType != BalooResource.ResourceTypes.URL)
                {
                    txtResURL.DataBindings.Clear();
                    txtResURL.Text = "";
                } else {
                    txtResURL.Text = resource.Location;
                    SetTextBinding(txtResURL, resource, "Location");
                    radioResURL.Checked = true;
                }
                if (resource.ResourceType == BalooResource.ResourceTypes.CommandList)
                    radioResList.Checked = true;

                chkResDisable.DataBindings.Clear();
                chkResDisable.DataBindings.Add( "Checked", resource, "Disabled" );
                chkResHide.DataBindings.Clear();
                chkResHide.DataBindings.Add( "Checked", resource, "Hidden" );

                grpEditResource.Visible = true;

            }
            else grpEditResource.Visible = false;
        }

        /// <summary>
        /// Deletes the resource (after showing a confirmation box).
        /// </summary>
        private void btnResDelete_Click(object sender, EventArgs e)
        {
            if (lstResources.SelectedItem != null && MessageBox.Show("Are you sure you want to delete this resource?", "Delete resource?", MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) == DialogResult.Yes)
            {
                broadcast.settings.Resources.Remove(lstResources.SelectedItem);
                ResourcesUpdate();
                lstResources.SelectedItem = null;
                grpEditResource.Visible = false;
            }
        }
        #endregion

        #region Editing resources
        /// <summary>
        /// When the name is updated, update the resource list so the new name is shown.
        /// </summary>
        private void txtResName_TextChanged(object sender, EventArgs e)
        {
            if (lstResources.SelectedItem != null)
            {
                ((BalooResource)lstResources.SelectedItem).Name = txtResName.Text;
                ResourcesUpdate();
            }
        }

        /// <summary>
        /// Checks if there are any spaces in the command, and if there is, remove them.
        /// </summary>
        private void txtResCommand_TextChanged(object sender, EventArgs e)
        {
            if (txtResCommand.Text.Contains(" "))
                txtResCommand.Text = txtResCommand.Text.Replace(" ", "");
        }

        /// <summary>
        /// Changes the DataBindings when the type of resource is changed.
        /// </summary>
        private void ResourceTypeChanged(object sender, EventArgs e)
        {
            BalooResource resource = (BalooResource)lstResources.SelectedItem;
            if (!radioResFile.Checked)
                txtResFile.DataBindings.Clear();
            else
            {
                resource.Location = txtResFile.Text;
                SetTextBinding(txtResFile, resource, "Location");
                radioResFile.Checked = true;
                resource.ResourceType = BalooResource.ResourceTypes.File;
            }
            if (!radioResFolder.Checked)
                txtResFolder.DataBindings.Clear();
            else
            {
                resource.Location = txtResFolder.Text;
                SetTextBinding(txtResFolder, resource, "Location");
                radioResFolder.Checked = true;
                resource.ResourceType = BalooResource.ResourceTypes.Folder;
            }
            if (!radioResURL.Checked)
                txtResURL.DataBindings.Clear();
            else
            {
                resource.Location = txtResURL.Text;
                SetTextBinding(txtResURL, resource, "Location");
                radioResURL.Checked = true;
                resource.ResourceType = BalooResource.ResourceTypes.URL;
            }
            if (radioResList.Checked)
                resource.ResourceType = BalooResource.ResourceTypes.CommandList;
            btnResFile.Enabled = radioResFile.Checked;
            txtResFile.Enabled = radioResFile.Checked;
            btnResFolder.Enabled = radioResFolder.Checked;
            txtResFolder.Enabled = radioResFolder.Checked;
            txtResURL.Enabled = radioResURL.Checked;
        }
        
        /// <summary>
        /// Shows the "Select file dialog and puts the result in the file text
        /// field.
        /// </summary>
        private void btnResFile_Click(object sender, EventArgs e)
        {
            openFileDialog.FileName = txtResFile.Text;
            openFileDialog.Filter = "Common mobile phone files (*.jpg;*.gif;*.txt;*.wav;*.mp3;*.vcf;*.vcs)|*.jpg;*.gif;*.txt;*.wav;*.mp3;*.vcf;*.vcs|All files (*.*)|*.*";
            if (openFileDialog.ShowDialog(this) == DialogResult.OK)
            {
                ((BalooResource)lstResources.SelectedItem).Location = openFileDialog.FileName;
                txtResFile.Text = openFileDialog.FileName;
            }
        }

        /// <summary>
        /// Shows the "Select folder" dialog and puts the result in the folder
        /// text field.
        /// </summary>
        private void btnResFolder_Click(object sender, EventArgs e)
        {
            folderBrowserDialog.SelectedPath = txtResFolder.Text;
            if (folderBrowserDialog.ShowDialog(this) == DialogResult.OK)
            {
                ((BalooResource)lstResources.SelectedItem).Location = folderBrowserDialog.SelectedPath;
                txtResFolder.Text = folderBrowserDialog.SelectedPath;
            }
        }
        #endregion
    }
}
