/* BalooGUI
 * Author: Johan Ho
 */
using System;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Xml.Serialization;
using System.Data;
using MySQLDriverCS;

namespace Baloo
{
    /// <summary>
    /// This partial class contains the code for handling input from the GUI elements
    /// in Baloo's GUI, except for GUI elements related to resources. The code for
    /// those elements are in BalooGUI.Resources.cs.
    /// </summary>
    public partial class BalooGUI : Form
    {
        private BalooBroadcast broadcast;
        private BalooBluetooth bluetooth;
        private string settingsdirectory;

        /// <summary>
        /// Initializes BalooGUI, and loads settings from settings.xml if it's found.
        /// Also creates instances of BalooBroadcast and BalooBluetooth.
        /// </summary>
        public BalooGUI()
        {
            settingsdirectory = Environment.CurrentDirectory;
            InitializeComponent();
            broadcast = new BalooBroadcast(txtLogging);
            if (File.Exists("settings.xml"))
                LoadSettings("settings.xml");
            else
            {
                broadcast.settings = new BalooSettings();
                broadcast.settings.Resources.Add(new BalooResource());
                lstResources.DataSource = broadcast.settings.Resources;
                lstResources.DisplayMember = "Name";
                broadcast.settings.Resources.Clear();
                ResourcesUpdate();
                SetConfigurationBindings();
            }
            bluetooth = new BalooBluetooth(broadcast,listViewThreads);
            comboMessages.DataSource = BalooSettings.Messages;
        }

        #region Main GUI functions (saving/loading settings, exiting)
        /// <summary>
        /// Either starts the broadcast, or if the broadcast is running, stops it.
        /// Disables the most of the GUI so that settings can't be changed during
        /// a broadcast. (Enables it if the broadcast has been stopped.)
        /// </summary>
        private void btnStartBroadcast_Click(object sender, EventArgs e)
        {
            bool starting = (btnStartBroadcast.Text == "Start broadcast");
            if (starting)
            {
                tabConfig.Show();
                if (bluetooth.StartBroadcast())
                {
                    broadcast.LogEvent("Started broadcast");
                    btnStartBroadcast.Text = "Stop broadcast";
                }
                else starting = false;
            }
            else
            {
                bluetooth.StopBroadcast();
                broadcast.LogEvent("Stopped broadcast");
                btnStartBroadcast.Text = "Start broadcast";
            }
            grpBroadcast.Enabled = !starting;
            grpMySQL.Enabled = !starting;
            txtLogFile.Enabled = !starting;
            btnLogFile.Enabled = !starting;
            grpEditResource.Enabled = !starting;
            btnNewResource.Enabled = !starting;
        }

        /// <summary>
        /// Shows a confirmation box asking if the user really wants to exit.
        /// </summary>
        private void BalooGUI_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (MessageBox.Show("Are you sure you want to exit Baloo?", "Exit Baloo?", MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) == DialogResult.No)
                e.Cancel = true;
            else bluetooth.StopBroadcast();
        }

        /// <summary>Exits Baloo.</summary>
        private void btnExitBaloo_Click(object sender, EventArgs e)
        {   this.Close();   }

        /// <summary>
        /// This function is used to show messages. This way we can easily change
        /// the way messages are shown, so we can switch between showing in a
        /// message box or in the status bar.
        /// </summary>
        /// <param name="message">The message to be shown</param>
        private void ShowMessage(string message)
        {
            lblStatusMessage.Text = message;
            //MessageBox.Show(message);
        }

        /// <summary>
        /// Shows a dialog box for selecting which file to save to, and then uses
        /// XmlSerializer to save the settings.
        /// </summary>
        private void btnSaveSettings_Click(object sender, EventArgs e)
        {
            saveFileDialog.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*";
            saveFileDialog.FileName = "";
            saveFileDialog.InitialDirectory = settingsdirectory;
            if (saveFileDialog.ShowDialog(this) == DialogResult.OK)
            {
                try
                {
                    StreamWriter writer = new StreamWriter(saveFileDialog.FileName, false, Encoding.UTF8);
                    XmlSerializer serializer = new XmlSerializer(typeof(BalooSettings));
                    serializer.Serialize(writer, broadcast.settings);
                    writer.Close();
                    ShowMessage("Settings and resources saved to " + saveFileDialog.FileName);
                    settingsdirectory = saveFileDialog.FileName.Substring(0,saveFileDialog.FileName.LastIndexOf('\\'));
                }
                catch (Exception ioe)
                {
                    ShowMessage("Error saving settings and resources: " + ioe.Message);
                }
            }
        }

        /// <summary>
        /// Shows the file dialog for selecting a settings file, and loads the settings
        /// from that file.
        /// </summary>
        private void btnLoadSettings_Click(object sender, EventArgs e)
        {
            openFileDialog.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*";
            openFileDialog.FileName = "";
            openFileDialog.InitialDirectory = settingsdirectory;
            if (openFileDialog.ShowDialog(this) == DialogResult.OK)
            {
                LoadSettings(openFileDialog.FileName);
            }
        }

        /// <summary>
        /// Loads settings from the specified file using XmlSerializer.
        /// </summary>
        /// <param name="file"></param>
        private void LoadSettings(string file)
        {
            // Read settings from the file.
            try
            {
                FileStream fs = File.OpenRead(file);
                XmlSerializer serializer = new XmlSerializer(typeof(BalooSettings));
                broadcast.settings = (BalooSettings)serializer.Deserialize(fs);
                fs.Close();
            }
            catch (IOException ioe)
            {
                ShowMessage("Error loading file: " + ioe.Message);
                return;
            }
            // If settings were successfully loaded, update the data bindings.
            SetConfigurationBindings();
            ShowMessage("Settings and resources loaded.");
        }
        #endregion

        #region Configuration tab functions

        /// <summary>
        /// Helper function for clearing away any old bindings before setting the new binding
        /// for a text field.
        /// </summary>
        /// <param name="textbox">The textbox which should be bound.</param>
        /// <param name="object">The object to bind the textbox to.</param>
        /// <param name="property">The BalooSettings property to bind the textbox to.</param>
        private void SetTextBinding(TextBox textbox, object obj, string property)
        {
            textbox.DataBindings.Clear();
            textbox.DataBindings.Add("Text", obj, property); ;
        }

        /// <summary>
        /// This function sets all necessary databindings for the textfields, to make sure
        /// the settings are updated when the text field is changed.
        /// </summary>
        private void SetConfigurationBindings()
        {
            comboStatisticsType.DataSource = BalooStatistics.StatisticsDescriptions;
            SetTextBinding(txtBroadcastName, broadcast.settings, "Name");
            if (broadcast.settings.SaveIncomingToFolder != null && broadcast.settings.SaveIncomingToFolder.Length > 0)
            {
                SetTextBinding(txtSaveIncomingToFolder, broadcast.settings, "SaveIncomingToFolder");
                radioSaveFiles.Checked = true;
            }
            else
            {
                txtSaveIncomingToFolder.Text = "";
                radioDenyFiles.Checked = true;
            }
            SetTextBinding(txtLogFile, broadcast.settings, "LogFile");
            SetTextBinding(txtIncomingFilter, broadcast.settings, "IncomingFilter");
            SetTextBinding(txtMySQLServer, broadcast.settings, "MySQLServer");
            SetTextBinding(txtMySQLUsername, broadcast.settings, "MySQLUsername");
            SetTextBinding(txtMySQLPassword, broadcast.settings, "MySQLPassword");
            SetTextBinding(txtMySQLDatabase, broadcast.settings, "MySQLDatabase");

            if (broadcast.settings.Resources.Count == 0)
            {
                broadcast.settings.Resources.Add(new BalooResource());
                lstResources.DataSource = broadcast.settings.Resources;
                broadcast.settings.Resources.Clear();
            }
            else lstResources.DataSource = broadcast.settings.Resources;
            lstResources.DisplayMember = "Name";
            lstResources.SelectedItem = null;

            // Refresh the resources listbox
            ResourcesUpdate();
        }

        /// <summary>
        /// This function is used by the two radio buttons for selecting
        /// what to do with incoming files.
        /// The text field and button for selecting the folder for saving
        /// incoming files is disabled/enabled depending on which radio
        /// button was checked.
        /// </summary>
        private void IncomingFilesActionChanged(object sender, EventArgs e)
        {
            txtSaveIncomingToFolder.Enabled = radioSaveFiles.Checked;
            btnSaveIncomingToFolder.Enabled = radioSaveFiles.Checked;
            if (radioDenyFiles.Checked)
            {
                txtSaveIncomingToFolder.DataBindings.Clear();
                broadcast.settings.SaveIncomingToFolder = null;
            } else {
                if (txtSaveIncomingToFolder.Text != "")
                    broadcast.settings.SaveIncomingToFolder = txtSaveIncomingToFolder.Text;
                SetTextBinding(txtSaveIncomingToFolder, broadcast.settings, "SaveIncomingToFolder");
            }
        }

        /// <summary>
        /// Shows the "Select folder" dialog window and saves the result
        /// to the text field.
        /// </summary>
        private void btnSaveIncomingTo_Click(object sender, EventArgs e)
        {
            folderBrowserDialog.SelectedPath = txtSaveIncomingToFolder.Text;
            if (folderBrowserDialog.ShowDialog(this) == DialogResult.OK)
            {
                txtSaveIncomingToFolder.Text = folderBrowserDialog.SelectedPath;
                broadcast.settings.SaveIncomingToFolder = folderBrowserDialog.SelectedPath;
            }
        }

        /// <summary>
        /// Shows the current message for the selected message type.
        /// </summary>
        private void comboMessages_SelectedIndexChanged(object sender, EventArgs e)
        {
            switch (comboMessages.SelectedIndex)
            {
                case BalooSettings.MSG_FILENOTFOUND:
                    SetTextBinding(txtMessageText, broadcast.settings, "FileNotFound");
                    break;
                case BalooSettings.MSG_INCOMINGFILESAVED:
                    SetTextBinding(txtMessageText, broadcast.settings, "IncomingFileSaved");
                    break;
                case BalooSettings.MSG_INCOMINGFILENOTACCEPTED:
                    SetTextBinding(txtMessageText, broadcast.settings, "IncomingFileNotAccepted");
                    break;
                case BalooSettings.MSG_RESOURCEUNAVAILABLE:
                    SetTextBinding(txtMessageText, broadcast.settings, "ResourceUnavailable");
                    break;
            }
        }

        /// <summary>
        /// Shows the "Choose file" dialog and writes the result in the
        /// logfile text field.
        /// </summary>
        private void btnLogFile_Click(object sender, EventArgs e)
        {
            saveFileDialog.Filter = "Log files (*.log)|*.log|All files (*.*)|*.*";
            saveFileDialog.FileName = txtLogFile.Text;
            saveFileDialog.InitialDirectory = txtLogFile.Text;
            saveFileDialog.OverwritePrompt = false;
            if (saveFileDialog.ShowDialog(this) == DialogResult.OK)
            {
                txtLogFile.Text = saveFileDialog.FileName;
                broadcast.settings.LogFile = saveFileDialog.FileName;
            }
            saveFileDialog.OverwritePrompt = true;
        }
        #endregion

        #region Statistics tab
        /// <summary>
        /// Generates statistics of the type that was selected.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void comboStatisticsType_SelectedIndexChanged(object sender, EventArgs e)
        {
            GenerateStatistics();
        }

        /// <summary>
        /// Helper function for generating statistics.
        /// </summary>
        private void GenerateStatistics()
        {
            UseWaitCursor = true;
            ShowMessage(broadcast.GenerateStatistics(
                dataGridStatistics,
                (BalooStatistics.StatisticsTypes)comboStatisticsType.SelectedIndex
            ));
            UseWaitCursor = false;
        }

        /// <summary>
        /// Shows a confirmation box, before deleting events older than the
        /// date selected in the DateTime selector.
        /// </summary>
        private void btnDeleteStatistics_Click(object sender, EventArgs e)
        {
            int result = broadcast.DeleteEventsBefore(dateTimeDeleteStatistics.Value);
            GenerateStatistics();
            if (result == 0)
                ShowMessage( "No events deleted." );
            else if (result == 1)
                ShowMessage( "One event deleted." );
            else if (result > 1)
                ShowMessage( result + " events deleted." );
            else ShowMessage( "Unknown error while deleting events!" );
        }
        #endregion

        /// <summary>
        /// Refreshes the statistics when the statistics tab is shown.
        /// </summary>
        private void tabStatistics_Enter(object sender, EventArgs e)
        {
            GenerateStatistics();
        }
    }
}