/* BalooStatistics
 * Author: Johan Ho
 */
using System;
using System.Data;
using System.Windows.Forms;
using MySQLDriverCS;

namespace Baloo
{
    /// <summary>
    /// This class is used for logging events to the given MySQL database.
    /// The table used is created automatically if it doesn't exist.
    /// </summary>
    class BalooStatistics
    {
        private string server;
        private string username;
        private string password;
        private string database;
        private MySQLConnection connection;
        private MySQLCommand insertevent;
        private DataGridViewColumn[] alleventscolumns, statisticscolumns;

        /* Types of statistics that can be shown. The string array is used by the
         * ComboBox in the GUI. Reordering won't break anything, but make sure
         * the order of types is the same in the enum and string array!
         */
        public enum StatisticsTypes {
            AllEvents, Hourly, Daily, CommandsPopularity
        }
        public static string[] StatisticsDescriptions
        {
            get {
                return new string[] {
                    "Show all events",
                    "Hourly statistics",
                    "Daily statistics",
                    "Commands popularity"
                };
            }
        }

        /// <summary>
        /// Initializes a new instance of BalooStatistics. This also creates the
        /// table needed (if it doesn't exist), 
        /// </summary>
        /// <param name="server"></param>
        /// <param name="username"></param>
        /// <param name="password"></param>
        /// <param name="database"></param>
        public BalooStatistics(string server, string username, string password, string database)
        {
            this.server = server;
            this.username = username;
            this.password = password;
            this.database = database;
            CreateTables();

            insertevent = new MySQLCommand( @"INSERT INTO event VALUES(
                '', NOW(), @command, @parameters, @reply, @hash
            )");

            /* The following columns are used for showing statistics in the
             * GenerateStatistics function. They're initialized here so they
             * don't need to be initialized every time statistics are shown. */
            #region Premaking columns for the datagrid
            alleventscolumns = new DataGridViewColumn[6];
            alleventscolumns[0] = new DataGridViewTextBoxColumn();
            alleventscolumns[0].HeaderText = "id";
            alleventscolumns[1] = new DataGridViewTextBoxColumn();
            alleventscolumns[1].HeaderText = "datetime";
            alleventscolumns[2] = new DataGridViewTextBoxColumn();
            alleventscolumns[2].HeaderText = "command";
            alleventscolumns[3] = new DataGridViewTextBoxColumn();
            alleventscolumns[3].HeaderText = "parameters";
            alleventscolumns[4] = new DataGridViewTextBoxColumn();
            alleventscolumns[4].HeaderText = "reply";
            alleventscolumns[5] = new DataGridViewTextBoxColumn();
            alleventscolumns[5].HeaderText = "hash";

            statisticscolumns = new DataGridViewColumn[2];
            statisticscolumns[0] = new DataGridViewTextBoxColumn();
            statisticscolumns[0].HeaderText = "hits";
            statisticscolumns[1] = new DataGridViewTextBoxColumn();
            statisticscolumns[1].HeaderText = "uniquehits";
            #endregion
        }

        /// <summary>
        /// Connects to the MySQL server using variables that were set when
        /// initializing this BalooStatistics object.
        /// </summary>
        /// <returns>True if connected, false if an error occured or if
        /// the connection variables aren't set.</returns>
        private bool Connect()
        {
            if (String.IsNullOrEmpty(server) || String.IsNullOrEmpty(database))
                return false;
            connection = new MySQLConnection(
                new MySQLConnectionString(server, database, username, password ).AsString
            );
            try {
                connection.Open();
                return true;
            } catch {
                return false;
            }
        }

        /// <summary>
        /// Disconnects from the MySQL server.
        /// </summary>
        private void Disconnect()
        {
            connection.Close();
        }

        /// <summary>
        /// Tests if the MySQL connection works.
        /// </summary>
        /// <returns>True if the connection to the MySQL server works, false if not.</returns>
        public bool ConnectionOK()
        {
            bool result = Connect();
            if (result)
                Disconnect();
            return result;
        }

        #region MySQL database manipulation functions
        /// <summary>
        /// Creates the table used for storing events, if it doesn't already exist.
        /// </summary>
        private void CreateTables()
        {
            if (Connect())
            {
                MySQLCommand cmd = new MySQLCommand(
                    "CREATE TABLE IF NOT EXISTS event (" +
                        "id INT NOT NULL AUTO_INCREMENT UNIQUE, " +
                        "datetime DATETIME NOT NULL, " +
                        "command VARCHAR(64) NOT NULL, " +
                        "parameters VARCHAR(255), " +
                        "reply TEXT NOT NULL, " +
                        "hash INT," +
                    "PRIMARY KEY (id) )", connection);
                int result = cmd.ExecuteNonQuery();
                //MessageBox.Show("Result was: " + result);
                Disconnect();
            }
        }

        /// <summary>
        /// Adds an event to the statistics.
        /// </summary>
        /// <param name="command">The command that was called.</param>
        /// <param name="parameters">The parameters passed witht he command, if any.</param>
        /// <param name="reply">The reply that was sent to the user.</param>
        /// <param name="hash">The hash of the Bluetooth ID of the user.</param>
        public void AddEvent(string command, string parameters, string reply, int hash)
        {
            if (Connect())
            {
                insertevent.Connection = connection;

                //Parameters are created and added to the insertevent command object
                MySQLParameter cmdparam = new MySQLParameter("@command", DbType.String);
                cmdparam.Value = command;
                MySQLParameter parametersparam = new MySQLParameter("@parameters", DbType.String);
                parametersparam.Value = parameters;
                MySQLParameter replyparam = new MySQLParameter("@reply", DbType.String);
                replyparam.Value = reply;
                MySQLParameter hashparam = new MySQLParameter("@hash", DbType.Int32);
                hashparam.Value = hash;

                insertevent.Parameters.Clear();
                insertevent.Parameters.Add(cmdparam);
                insertevent.Parameters.Add(parametersparam);
                insertevent.Parameters.Add(replyparam);
                insertevent.Parameters.Add(hashparam);

                int result = insertevent.ExecuteNonQuery();

                Disconnect();
            }
        }

        public int DeleteEventsBefore(DateTime date)
        {
            if (Connect())
            {
                MySQLCommand cmd = (MySQLCommand)connection.CreateCommand();
                cmd.CommandText = "DELETE FROM event WHERE datetime < @date";
                MySQLParameter parameter = new MySQLParameter("@date", DbType.Date);
                parameter.Value = date;
                cmd.Parameters.Add(parameter);
                int result = cmd.ExecuteNonQuery();
                connection.Close();
                return result;
            }
            return 0;
        }
        #endregion

        #region Statistics generation (used with GUI)
        /// <summary>
        /// Generates statistics and shows it in the given datagrid component.
        /// This function may need to be extended to allow for showing only statistics
        /// between specific dates/times, or for showing only a page of results at a
        /// time.
        /// </summary>
        /// <param name="dataGridStatistics">The datagrid to show the statistics in.</param>
        /// <param name="StatisticsTypes">The type of statistics to be shown.</param>
        public string GenerateStatistics(DataGridView dataGridStatistics, StatisticsTypes StatisticsTypes)
        {
            if (Connect())
            {
                try
                {
                    dataGridStatistics.Columns.Clear();
                    switch (StatisticsTypes)
                    {
                        case StatisticsTypes.AllEvents:
                            GenerateStatisticsAllEvents(dataGridStatistics);
                            break;
                        case StatisticsTypes.Hourly:
                            GenerateStatisticsHourly(dataGridStatistics);
                            break;
                        case StatisticsTypes.Daily:
                            GenerateStatisticsDaily(dataGridStatistics);
                            break;
                        case StatisticsTypes.CommandsPopularity:
                            GenerateStatisticsCommandsPopularity(dataGridStatistics);
                            break;
                    }
                    connection.Close();
                }
                catch (Exception error)
                {
                    return "An error occurred while generating the statistics: " + error.Message;
                }
                return "Statistics generated";
            }
            else return "Couldn't connect to MySQL server. Check your MySQL settings.";
        }

        /// <summary>
        /// Helper function for generating a list of all events in the database.
        /// </summary>
        /// <param name="dataGridStatistics">The datagrid used for showing the statistics.</param>
        private void GenerateStatisticsAllEvents(DataGridView dataGridStatistics)
        {
            dataGridStatistics.Columns.AddRange(alleventscolumns);
            MySQLCommand cmd = (MySQLCommand)connection.CreateCommand();
            cmd.CommandText = "SELECT * FROM event";
            MySQLDataReader reader = (MySQLDataReader)cmd.ExecuteReader();
            while (reader.Read())
            {
                dataGridStatistics.Rows.Add(new object[] {
                                reader.GetInt32(0),
                                reader.GetDateTime(1),
                                reader.GetString(2),
                                reader.GetString(3),
                                reader.GetString(4),
                                ( !reader.IsDBNull(5)? reader.GetInt32(5) : 0 )
                            });
            }
        }

        /// <summary>
        /// Helper function for showing hits and unique hits for each hour in each date
        /// which contains events.
        /// </summary>
        /// <param name="dataGridStatistics">The datagrid used for showing the statistics.</param>
        private void GenerateStatisticsHourly(DataGridView dataGridStatistics)
        {
            DataGridViewTextBoxColumn textboxcolumn;
            textboxcolumn = new DataGridViewTextBoxColumn();
            textboxcolumn.Name = "datehour";
            dataGridStatistics.Columns.Add(textboxcolumn);
            dataGridStatistics.Columns.AddRange(statisticscolumns);
            MySQLCommand cmd = (MySQLCommand)connection.CreateCommand();
            cmd.CommandText = "SELECT CONCAT(SUBSTRING(datetime,1,10), ' h: ', HOUR(datetime)) AS datehour, COUNT(*) AS hits,COUNT(DISTINCT hash) AS uniquehits FROM `event` GROUP BY datehour";
            MySQLDataReader reader = (MySQLDataReader)cmd.ExecuteReader();
            while (reader.Read())
            {
                dataGridStatistics.Rows.Add(new object[] {
                                reader.GetString(0),
                                reader.GetInt32(1),
                                reader.GetInt32(2)
                            });
            }
        }

        /// <summary>
        /// Helper function for showing hits and unique hits per day.
        /// </summary>
        /// <param name="dataGridStatistics">The datagrid used for showing the statistics.</param>
        private void GenerateStatisticsDaily(DataGridView dataGridStatistics)
        {
            MySQLCommand cmd = (MySQLCommand)connection.CreateCommand();
            DataGridViewTextBoxColumn textboxcolumn;
            textboxcolumn = new DataGridViewTextBoxColumn();
            textboxcolumn.Name = "date";
            dataGridStatistics.Columns.Add(textboxcolumn);
            textboxcolumn = new DataGridViewTextBoxColumn();
            dataGridStatistics.Columns.AddRange(statisticscolumns);
            cmd.CommandText = "SELECT SUBSTRING(datetime,1,10) AS date, COUNT(*) AS hits,COUNT(DISTINCT hash) AS uniquehits FROM `event` GROUP BY date";
            MySQLDataReader reader = (MySQLDataReader)cmd.ExecuteReader();
            while (reader.Read())
            {
                dataGridStatistics.Rows.Add(new object[] {
                                reader.GetString(0),
                                reader.GetInt32(1),
                                reader.GetInt32(2)
                            });
            }
        }

        /// <summary>
        /// Helper function for generating statistics showing commands and their hits
        /// and unique hits.
        /// </summary>
        /// <param name="dataGridStatistics">The datagrid used for showing the statistics.</param>
        private void GenerateStatisticsCommandsPopularity(DataGridView dataGridStatistics)
        {
            DataGridViewTextBoxColumn textboxcolumn;
            textboxcolumn = new DataGridViewTextBoxColumn();
            textboxcolumn.Name = "command";
            dataGridStatistics.Columns.Add(textboxcolumn);
            textboxcolumn = new DataGridViewTextBoxColumn();
            textboxcolumn.Name = "requests";
            dataGridStatistics.Columns.Add(textboxcolumn);
            textboxcolumn = new DataGridViewTextBoxColumn();
            textboxcolumn.Name = "uniquerequests";
            dataGridStatistics.Columns.Add(textboxcolumn);
            MySQLCommand cmd = (MySQLCommand)connection.CreateCommand();
            cmd.CommandText = "SELECT command, COUNT(*) AS requests,COUNT(DISTINCT hash) AS uniquerequests FROM `event` GROUP BY command";
            MySQLDataReader reader = (MySQLDataReader)cmd.ExecuteReader();
            while (reader.Read())
            {
                dataGridStatistics.Rows.Add(new object[] {
                                reader.GetString(0),
                                reader.GetInt32(1),
                                reader.GetInt32(2)
                            });
            }
        }
        #endregion
    }
}
